==> em_product-0.2.12-py3-none-any.whl: em_product/resources/mongo_pipelines.py <==
# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html


# useful for handling different item types with a single interface
from datetime import datetime, timedelta
import logging
import json
import time

from pymongo import MongoClient, UpdateOne
from pymongo.errors import (
    BulkWriteError,
    ConnectionFailure,
    OperationFailure,
    ServerSelectionTimeoutError,
)


class MongoPipeline:
    BUFFER_SIZE = 1000
    COLLECTION = None
    MAX_RETRY = 3

    def __init__(self, mongo_uri, mongo_db):
        self.client = MongoClient(mongo_uri)
        self.mongo_db = self.client[mongo_db]
        self.items_buffer = []

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            mongo_uri=crawler.settings.get("MONGO_URI"),
            mongo_db=crawler.settings.get("MONGO_DATABASE"),
        )

    def save_item(self, item):
        self.items_buffer.append(UpdateOne({"_id": item.pop("_id")}, {"$set": item}, upsert=True))
        if len(self.items_buffer) >= self.BUFFER_SIZE:
            self.send_items()
            self.items_buffer.clear()

    def send_items(self):
        retry_left_count = self.MAX_RETRY
        while retry_left_count > 0:
            try:
                self.mongo_db[self.COLLECTION].bulk_write(self.items_buffer)
                logging.info(f"[Items Saved]: {self.items_buffer}")
                break
            except BulkWriteError as bwe:
                for error in bwe.details["writeErrors"]:
                    logging.exception(f"{error['errmsg']}")
                break

            except OperationFailure as of:
                logging.exception(f"{str(of)}")
                break

            except (ConnectionFailure, ServerSelectionTimeoutError) as e:
                time.sleep(2 ** (self.MAX_RETRY - retry_left_count))
                retry_left_count = retry_left_count - 1

            except Exception as e:
                raise e

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            self.save_item(item)
            logging.debug("Item has been sent to Mongo DB %s", self.COLLECTION)

        return item

    def close_spider(self, spider):
        if len(self.items_buffer):
            self.send_items()


def is_expired(product):
    if not product.get("date"):
        return True
    try:
        if isinstance(product["date"], str):
            date_obj = datetime.strptime(product["date"], "%Y-%m-%dT%H:%M:%S")
            return datetime.now() - date_obj > timedelta(weeks=1)
        elif isinstance(product["date"], datetime):
            return datetime.now() - product["date"] > timedelta(weeks=1)
        else:
            return True

    except Exception as e:
        return True


class MongoProductPipeline(MongoPipeline):
    BUFFER_SIZE = 4000
    COLLECTION = None
    limit = None

    urls = {}

    def load_product_urls(self):
        for doc in self.mongo_db["product_urls"].find({}, batch_size=1000):
            if self.limit is not None:
                if len(self.urls) > self.limit:
                    for _, url in self.urls.items():
                        yield url
                    break
            if len(self.urls) < self.BUFFER_SIZE:
                self.urls[doc["_id"]] = doc["url"]
                continue

            products = self.mongo_db[self.COLLECTION].find({"_id": {"$in": list(self.urls.keys())}})
            for product in products:
                if not is_expired(product):
                    self.urls.pop(product["_id"], None)

            for _, url in self.urls.items():
                yield url
            self.urls.clear()

    def open_spider(self, spider):
        if self.COLLECTION is None:
            raise Exception("no collection is provided")

        for url in self.load_product_urls():
            spider.start_urls.append(url)


class ElasticSearchPipeline:
    ELASTICSEARCH_BUFFER_LENGTH = 500
    ELASTICSEARCH_INDEX = None
    FORCE = False

    def __init__(self, settings, id_key=None):
        self.id_key = id_key
        self.max_retry = 3
        self.init_es_client(settings)
        self.items_buffer = []

    def init_es_client(self, settings):
        es_timeout = settings.get("ELASTICSEARCH_TIMEOUT", 60)
        es_servers = settings.get("ELASTICSEARCH_SERVERS")
        es_servers = es_servers if isinstance(es_servers, list) else [es_servers]

        es_settings = dict()
        es_settings["hosts"] = es_servers
        es_settings["timeout"] = es_timeout

        if settings.get("ELASTICSEARCH_USERNAME") and settings.get("ELASTICSEARCH_PASSWORD"):
            es_settings["http_auth"] = (
                settings["ELASTICSEARCH_USERNAME"],
                settings["ELASTICSEARCH_PASSWORD"],
            )

        self.es = Elasticsearch(**es_settings)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def update_item(self, item):
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        update_action = dict()
        if "_id" in item:
            update_action["_id"] = item.pop("_id")
        else:
            if self.id_key is not None and self.id_key in item:
                update_action["_id"] = item[self.id_key]
        update_action.update(
            {
                "_index": index_name,
                "_type": "_doc",
                "_op_type": "update",
                "doc_as_upsert": True,
                "doc": dict(item),
            }
        )

        self.items_buffer.append(update_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def save_item(self, item, action="create"):
        item.pop("_id", None)
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        index_action = {
            "_index": index_name,
            "_type": "_doc",
            "_op_type": action,
            "_source": dict(item),
        }

        if self.id_key is not None and self.id_key in item:
            index_action["_id"] = item[self.id_key]

        self.items_buffer.append(index_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def create_item(self, item):
        item.pop("_id", None)
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        index_action = {
            "_index": index_name,
            "_type": "_doc",
            "_op_type": "create",
            "_source": dict(item),
        }

        if self.id_key is not None and self.id_key in item:
            index_action["_id"] = item[self.id_key]

        self.items_buffer.append(index_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def index_item(self, item):
        item.pop("_id", None)
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        index_action = {
            "_index": index_name,
            "_type": "_doc",
            "_op_type": "index",
            "_source": dict(item),
        }

        if self.id_key is not None and self.id_key in item:
            index_action["_id"] = item[self.id_key]

        self.items_buffer.append(index_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def send_items(self):
        retry = self.max_retry
        while retry > 0:
            try:
                res = helpers.bulk(self.es, self.items_buffer)
                logging.info(f"[Items Saved]: {self.items_buffer}")
                break
            except (ConnectionTimeout, ConnectionError, SSLError, TransportError):
                time.sleep(retry)
                retry -= 1
                continue
            except RequestError as e:
                break
            except BulkIndexError as e:
                for error in e.errors:
                    for _, value in error.items():
                        try:
                            logging.exception(value["_id"])
                        except Exception as e:
                            pass
                break
            except Exception as e:
                raise e

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            self.save_item(item)
            logging.debug("Item has been sent to Mongo DB %s", self.ELASTICSEARCH_INDEX)

        return item

    def close_spider(self, spider):
        if len(self.items_buffer):
            self.send_items()

==> em_product-0.2.14-py3-none-any.whl: em_product/resources/mongo_pipelines.py <==
# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html


# useful for handling different item types with a single interface
from datetime import datetime, timedelta
import logging
import json
import time

from pymongo import MongoClient, UpdateOne
from pymongo.errors import (
    BulkWriteError,
    ConnectionFailure,
    OperationFailure,
    ServerSelectionTimeoutError,
)


class MongoPipeline:
    BUFFER_SIZE = 1000
    COLLECTION = None
    MAX_RETRY = 3
    FORCE = None

    def __init__(self, mongo_uri, mongo_db):
        self.client = MongoClient(mongo_uri)
        self.mongo_db = self.client[mongo_db]
        self.items_buffer = []

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            mongo_uri=crawler.settings.get("MONGO_URI"),
            mongo_db=crawler.settings.get("MONGO_DATABASE"),
        )

    def save_item(self, item):
        self.items_buffer.append(UpdateOne({"_id": item.pop("_id")}, {"$set": item}, upsert=True))
        if len(self.items_buffer) >= self.BUFFER_SIZE:
            self.send_items()
            self.items_buffer.clear()

    def send_items(self):
        retry_left_count = self.MAX_RETRY
        while retry_left_count > 0:
            try:
                self.mongo_db[self.COLLECTION].bulk_write(self.items_buffer)
                logging.info(f"[Items Saved]: {self.items_buffer}")
                break
            except BulkWriteError as bwe:
                for error in bwe.details["writeErrors"]:
                    logging.exception(f"{error['errmsg']}")
                break

            except OperationFailure as of:
                logging.exception(f"{str(of)}")
                break

            except (ConnectionFailure, ServerSelectionTimeoutError) as e:
                time.sleep(2 ** (self.MAX_RETRY - retry_left_count))
                retry_left_count = retry_left_count - 1

            except Exception as e:
                raise e

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            self.save_item(item)
            logging.debug("Item has been sent to Mongo DB %s", self.COLLECTION)

        return item

    def close_spider(self, spider):
        if len(self.items_buffer):
            self.send_items()


def is_expired(product):
    if not product.get("date"):
        return True
    try:
        if isinstance(product["date"], str):
            date_obj = datetime.strptime(product["date"], "%Y-%m-%dT%H:%M:%S")
            return datetime.now() - date_obj > timedelta(weeks=1)
        elif isinstance(product["date"], datetime):
            return datetime.now() - product["date"] > timedelta(weeks=1)
        else:
            return True

    except Exception as e:
        return True


class MongoProductPipeline(MongoPipeline):
    BUFFER_SIZE = 4000
    COLLECTION = None
    limit = None

    urls = {}

    def load_product_urls(self):
        for doc in self.mongo_db["product_urls"].find({}, batch_size=1000):
            if self.limit is not None:
                if len(self.urls) > self.limit:
                    for _, url in self.urls.items():
                        yield url
                    break
            if len(self.urls) < self.BUFFER_SIZE:
                self.urls[doc["_id"]] = doc["url"]
                continue

            if self.FORCE is None:
                products = self.mongo_db[self.COLLECTION].find(
                    {"_id": {"$in": list(self.urls.keys())}}
                )
                for product in products:
                    if not is_expired(product):
                        self.urls.pop(product["_id"], None)

            for _, url in self.urls.items():
                yield url
            self.urls.clear()

    def open_spider(self, spider):
        if self.COLLECTION is None:
            raise Exception("no collection is provided")

        for url in self.load_product_urls():
            spider.start_urls.append(url)


class ElasticSearchPipeline:
    ELASTICSEARCH_BUFFER_LENGTH = 500
    ELASTICSEARCH_INDEX = None
    FORCE = False

    def __init__(self, settings, id_key=None):
        self.id_key = id_key
        self.max_retry = 3
        self.init_es_client(settings)
        self.items_buffer = []

    def init_es_client(self, settings):
        es_timeout = settings.get("ELASTICSEARCH_TIMEOUT", 60)
        es_servers = settings.get("ELASTICSEARCH_SERVERS")
        es_servers = es_servers if isinstance(es_servers, list) else [es_servers]

        es_settings = dict()
        es_settings["hosts"] = es_servers
        es_settings["timeout"] = es_timeout

        if settings.get("ELASTICSEARCH_USERNAME") and settings.get("ELASTICSEARCH_PASSWORD"):
            es_settings["http_auth"] = (
                settings["ELASTICSEARCH_USERNAME"],
                settings["ELASTICSEARCH_PASSWORD"],
            )

        self.es = Elasticsearch(**es_settings)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def update_item(self, item):
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        update_action = dict()
        if "_id" in item:
            update_action["_id"] = item.pop("_id")
        else:
            if self.id_key is not None and self.id_key in item:
                update_action["_id"] = item[self.id_key]
        update_action.update(
            {
                "_index": index_name,
                "_type": "_doc",
                "_op_type": "update",
                "doc_as_upsert": True,
                "doc": dict(item),
            }
        )

        self.items_buffer.append(update_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def save_item(self, item, action="create"):
        item.pop("_id", None)
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        index_action = {
            "_index": index_name,
            "_type": "_doc",
            "_op_type": action,
            "_source": dict(item),
        }

        if self.id_key is not None and self.id_key in item:
            index_action["_id"] = item[self.id_key]

        self.items_buffer.append(index_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def create_item(self, item):
        item.pop("_id", None)
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        index_action = {
            "_index": index_name,
            "_type": "_doc",
            "_op_type": "create",
            "_source": dict(item),
        }

        if self.id_key is not None and self.id_key in item:
            index_action["_id"] = item[self.id_key]

        self.items_buffer.append(index_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def index_item(self, item):
        item.pop("_id", None)
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        index_action = {
            "_index": index_name,
            "_type": "_doc",
            "_op_type": "index",
            "_source": dict(item),
        }

        if self.id_key is not None and self.id_key in item:
            index_action["_id"] = item[self.id_key]

        self.items_buffer.append(index_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def send_items(self):
        retry = self.max_retry
        while retry > 0:
            try:
                res = helpers.bulk(self.es, self.items_buffer)
                logging.info(f"[Items Saved]: {self.items_buffer}")
                break
            except (ConnectionTimeout, ConnectionError, SSLError, TransportError):
                time.sleep(retry)
                retry -= 1
                continue
            except RequestError as e:
                break
            except BulkIndexError as e:
                for error in e.errors:
                    for _, value in error.items():
                        try:
                            logging.exception(value["_id"])
                        except Exception as e:
                            pass
                break
            except Exception as e:
                raise e

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            self.save_item(item)
            logging.debug("Item has been sent to Mongo DB %s", self.ELASTICSEARCH_INDEX)

        return item

    def close_spider(self, spider):
        if len(self.items_buffer):
            self.send_items()

        if self.client:
            self.client.close()
